#!/usr/bin/env python
# stilt_tools.py

"""
Author : W. He
Adapted by Ingrid Super, last revisions 14-8-2018

Revision History:
Basing on Wouter's codes, replace the TM5 model with the STILT model, April 2015.

This module holds specific functions needed to use the STILT model within the data assimilation shell. It uses the information
from the DA system in combination with the generic stilt.rc files.

The STILT model is now controlled by a python subprocess. This subprocess consists of an MPI wrapper (written in C) that spawns
a large number ( N= nmembers) of STILT model instances under mpirun, and waits for them all to finish.

#The design of the system assumes that the stilt function are executed using R codes, and is residing in a
#directory specified by the ${RUNDIR} of a stilt rc-file. This stilt rc-file name is taken from the data assimilation rc-file. Thus,

"""
import shutil
import os
import sys
import logging
import datetime as dtm
import numpy as np
from numpy import array, logical_and
from string import join
import glob
import da.tools.io4 as io
import subprocess

import da.tools.rc as rc
from da.tools.general import create_dirs, to_datetime
from da.baseclasses.observationoperator import ObservationOperator

# global constants, which will be used in the following classes
identifier = 'WRF-STILT'
version = '1.0'

################### Begin Class STILT ###################

def dumfunc(arg):
    return STILTObservationOperator.OPS_run(*arg)

class STILTObservationOperator(object):

    def __init__(self, filename, dacycle=None):   #only the filename used to specify the location of the stavector file for wrf-stilt runs
        """ The instance of an STILTObservationOperator is application dependent """
        self.ID = identifier    # the identifier gives the model name
        self.version = version       # the model version used
        self.restart_filelist = []
        self.output_filelist = []
        self.outputdir = None # Needed for opening the samples.nc files created

        logging.info('Observation Operator initialized: %s (%s)' % (self.ID, self.version))

        self.load_rc(filename)   # load the specified rc-file
        
        if dacycle != None:
            self.dacycle = dacycle
        else:
            self.dacycle = {}

        self.startdate=None

    def get_initial_data(self):
        """ This method places all initial data needed by an ObservationOperator in the proper folder for the model """

    def setup(self, dacycle):
        """ Execute all steps needed to prepare the ObsOperator for use inside CTDAS, only done at the very first cycle normally """
        self.dacycle = dacycle
        self.outputdir = dacycle['dir.output']
        self.load_rc = dacycle['da.obsoperator.rc'] #stilt_footprint.rc
        self.obsfile = dacycle.dasystem['obs.input.id']
        self.obsdir = dacycle.dasystem['datadir']
        self.nrloc = int(dacycle.dasystem['obs.input.nr'])
        self.nrspc = int(dacycle.dasystem['obs.spec.nr'])
        self.nrcat = int(dacycle.dasystem['obs.cat.nr'])
        self.bgswitch = int(dacycle.dasystem['obs.bgswitch'])
        self.bgfile = dacycle.dasystem['obs.background']
        self.btime = int(dacycle.dasystem['run.backtime'])
        self.paramfile = dacycle.dasystem['emis.paramfile']
        
    def load_rc(self, name):
        """ 
        This method loads an OPS rc-file with settings for this simulation 
        """
        
        self.rcfile = rc.RcFile(name)
        self.model_settings = self.rcfile.values
        logging.debug('rc-file loaded successfully')

    def prepare_run(self,psdo,adv):
        """ Prepare the running of the actual forecast model, for example compile code """

        if psdo==0:
            # Define the name of the file that will contain the modeled output of each observation
            if adv==0:
                self.simulated_file = os.path.join(self.outputdir, 'samples_simulated.%s.nc' % self.dacycle['time.sample.stamp'])
            else:
                self.simulated_file = os.path.join(self.outputdir, 'samples_simulated.%s.adv.nc' % self.dacycle['time.sample.stamp'])
            self.forecast_nmembers = int(self.dacycle['da.optimizer.nmembers'])
            self.param_inputdir = self.dacycle['dir.input']
        
        #list all observation locations and species
        self.lat = []
        self.lon = []
        self.hgt = []
        self.obsnc = []
        self.mmass = []
        self.cr = []
        infile = os.path.join(self.obsdir, self.obsfile)    
        f = open(infile, 'r')
        lines = f.readlines()
        f.close()
        self.spname = []
        ct = 0
        for line in lines:
            dum=line.split(",")
            if dum[0]=='#':
                continue
            else:
                fln, lt, ln, hg, sp, mss, crr = (dum[1], float(dum[2]), float(dum[3]), float(dum[4]), dum[5], float(dum[6]), float(dum[7]))
                self.obsnc.append(fln)
                if sp in self.spname:
                    if sp == self.spname[0]:
                        self.lat.append(lt)
                        self.lon.append(ln)
                        self.hgt.append(hg)
                else:
                    self.spname.append(sp)
                    self.mmass.append(mss)
                    self.cr.append(crr)
                    if ct == 0:
                        self.lat.append(lt)
                        self.lon.append(ln)
                        self.hgt.append(hg)
                ct = ct + 1
                    
        #find out which model is to be used for which sector
        infile = os.path.join(self.obsdir, self.paramfile)
        f = open(infile, 'r')
        lines = f.readlines()
        f.close()
        self.ops_sector = []
        self.ops_id = []
        self.temporal_var = []
        ct = 0
        for line in lines:
            dum=line.split(",")
            if dum[0]=='#':
                continue
            else:
                id = int(dum[0])
                if id != 0:
                    self.ops_sector.append(ct)
                    self.ops_id.append(id)
                ct = ct + 1
                self.temporal_var.append(dum[11])
        
    def OPS_run(self,n,d,spc,psdo):
        """ This function prepares input files for OPS runs and reads the output"""
        
        #update the .inv file (model settings for OPS)     
        edhr = self.modelend.hour
        if edhr == 0:
            edhr = 24
            enddate=self.modelend-dtm.timedelta(0,3600)
        else:
            enddate=self.modelend
        
        invfile = self.model_settings['run.setupfile']
        file = open(invfile,'r')
        dum = file.readlines()
        file.close()
        invfile_new = os.path.join(self.dacycle['da.obsoperator.home'], 'run%03d%03d.inv'%(spc+1,n))
        file = open(invfile_new,'w')
        for line in dum:
            if 'component' in line:
                if self.spname[spc] == 'CO2' or self.spname[spc] == 'C14':
                    file.write('7')
                    file.write('\n')
                elif self.spname[spc] == 'CO':
                    file.write('8')
                    file.write('\n')
                elif self.spname[spc] == 'NOx':
                    file.write('2')
                    file.write('\n')
                elif self.spname[spc] == 'SO2':
                    file.write('1')
                    file.write('\n')
                else: #assume no chemistry or deposition for other species
                    file.write('8')
                    file.write('\n')
            elif 'date & hour begin' in line:
                file.write('%02d %02d %02d %02d\n'%(self.modelstart.year - 2000,self.modelstart.month,self.modelstart.day,1))
            elif 'date & hour end' in line:
                file.write('%02d %02d %02d %02d\n'%(enddate.year - 2000,enddate.month,enddate.day,edhr))
            elif 'meteofile' in line:
                file.write('./../../OPS_input/Meteo_ruwe_data/rg2mbk%02d\n'%(self.modelstart.year - 2000))
            elif 'averaging' in line:
                file.write(str(self.model_settings['run.avgopt']))
                file.write('\n')
            elif 'hourly (1) or annual (0)' in line:
                file.write(str(self.model_settings['run.emdt']))
                file.write('\n')
                ### currently CTDAS only supports run.emdt == 1, otherwise no time profiles can be included
            elif 'emission data file' in line:
                if self.model_settings['run.emdt']=='0':
                    file.write('./run%03d.brn'%(spc+1))
                    file.write('\n')
                else:
                    file.write('./dum.brn')
                    file.write('\n')
            elif 'hourly emission file' in line:
                if self.model_settings['run.emdt']=='0':
                    file.write('./dum.ems')
                    file.write('\n')
                else:
                    file.write('./run%03d%03d.ems'%(spc+1,n))
                    file.write('\n')
            else:
                file.write(line)
        file.close()
        
        #call function to create emission file with correct temporal variations for each ensemble member
        self.OPSsource(d=d,n=n,spc=spc,psdo=psdo)
        
        #set correct file paths and run OPS
        cwd = os.getcwd()
        os.chdir(self.dacycle['da.obsoperator.home'])
        outdir = os.path.join(self.model_settings['run.opsoutdir'],'%03d'%(spc+1))
        rptfile = self.model_settings['run.rptfile']
        newfile = 'run%03d.rpt'%(spc+1)
        shutil.copy(rptfile,os.path.join(self.dacycle['da.obsoperator.home'],newfile))
        FNULL = open(os.devnull, 'w')
        
        #extra check: compare final date output file with enddate in .inv file, since sometimes OPS files are prepared
        #but the run is not completed. Then try again!
        testdate = [0,0]
        while testdate != [enddate.day, edhr]:
            code = subprocess.call(["./run", "%03d%03d"%(spc+1,n)],stdout=FNULL)
            opsfile = os.path.join(outdir,'run%03d_%03d.plt'%(spc+1,n))
            shutil.copy('run%03d%03d.plt'%(spc+1,n),opsfile)
            out = np.loadtxt(opsfile,skiprows=2,delimiter=';',usecols=(0,1,2,3)).T
            testdate = [int(out[2,-1]),int(out[3,-1])]
        os.chdir(cwd)
        
        #read output, which contains output for one species but for all locations
        opsfile = os.path.join(outdir,'run%03d_%03d.plt'%(spc+1,n))
        out = np.loadtxt(opsfile,skiprows=2,delimiter=';',usecols=(0,4,11)).T
        loc = out[1]
        cnc = out[2]
        cncc = []
        for j in range(self.nrloc):
            dum = []
            for i in range(len(loc)):
                if loc[i]==j+1:
                    dum.append(cnc[i])
            cncc.append(dum)
        cncc = np.array(cncc)
        Vm = 24.45E-3
        if n==999: #pseudo-observations: read entire time series excluding first day of spin-up
            for j in range(self.nrloc):
                conc = cncc[j,24:]*self.cr[spc]*Vm/self.mmass[spc]
                self.conc_OPS[len(self.datlist)*spc+len(self.datlist)*self.nrspc*j:len(self.datlist)*spc+len(self.datlist)*self.nrspc*j+len(self.datlist)]=conc
        else: #inversion: read only final concentration for each locations
            conc = cncc[:,-1]*self.cr[spc]*Vm/self.mmass[spc]
                        
        return conc
        
    def OPSsource(self,d,n,spc,psdo):
        """This function creates hourly emission files for each OPS run"""

        dt=(self.modelend-self.modelstart).days*24+(self.modelend-self.modelstart).seconds/3600+1
        dtkey=(self.timefinishkey-self.timestartkey).days*24+(self.timefinishkey-self.timestartkey).seconds/3600+1
        #read in temporal profiles; always start at 1:00 a.m. because this is needed for OPS to run
        if psdo==0:
            idx=d+dtkey
            tempfile = os.path.join(self.model_settings['run.datadir'],'prior_temporal_%03d.nc'%n)
            tf = io.ct_read(tempfile, 'read')
            temp_profiles = []
            for s in self.ops_sector:
                f = tf.get_variable(self.temporal_var[s])[idx-dt:idx]
                temp_profiles.append(f)
            tf.close()
        elif psdo==1:
            tempfile = os.path.join(self.model_settings['run.datadir'],'prior_temporal_999.nc')
            tf = io.ct_read(tempfile, 'read')
            temp_profiles = []
            for s in self.ops_sector:
                f = tf.get_variable(self.temporal_var[s])
                temp_profiles.append(f)
            tf.close()
        temp_profiles = np.array(temp_profiles)
        
        #read total OPS emissions calculated with the emission model
        emisfile = os.path.join(self.model_settings['run.datadir'],'prior_spatial_%03d.nc'%n)
        f = io.ct_read(emisfile,method='read')
        OPS_tot=f.get_variable('OPStotals_%s'%(self.spname[spc]))
        f.close()
        
        #read prior emission file with fractions
        brnfile = os.path.join(self.dacycle['da.obsoperator.home'], 'run%s_prior.brn'%(self.spname[spc]))
        file = open(brnfile,'r')
        dum = file.readlines()
        file.close()
        header=[]
        data=[]
        ct=1
        for line in dum:
            if 'ssn' in line:
                header.append(line)
            else:
                xloc, yloc, hc, hgt, dim, sv, dv, cat, ar, sd = \
                    (int(line[4:12]), int(line[12:20]), float(line[30:37]), \
                    float(line[37:43]), int(line[43:50]), float(line[50:56]), int(line[56:60]), \
                    int(line[60:64]), int(line[64:68]), int(line[68:70]))
                emid = ct
                emis = float(line[20:30])
                data.append([emid,xloc,yloc,emis,hc,hgt,dim,sv,dv,cat,ar,sd])
                ct = ct+1
        data=np.array(data)
        
        # Open new emission file and fill with data. OPS always needs to start a run at 1:00 a.m. Since we want a specific footprint (self.btime)
        # we need to set all emissions prior to the start of the footprint to zero (flg = 1).
        brnfile_new = os.path.join(self.dacycle['da.obsoperator.home'], 'run%03d%03d.ems'%(spc+1,n))
        file = open(brnfile_new,'w')
        for t in range(dt):
            crdate = self.modelstart+dtm.timedelta(0,3600*t)
            if crdate < self.timestart and psdo==0:
                flg = 1
            else:
                flg = 0
            crhr = crdate.hour
            if crhr == 0:
                crhr = 24
                crdate = crdate - dtm.timedelta(0,3600)
            cryr = crdate.year-2000
            crmnt = crdate.month
            crdy = crdate.day
            for k in range(len(data)):
                if flg == 0:
                    for s in range(len(self.ops_sector)):
                        if data[k,8] == self.ops_id[s]:
                            emis = data[k,3]*OPS_tot[s]*temp_profiles[s,t]
                elif flg==1:
                    emis = 0.
                file.write('%4d%4d%4d%4d%6d%8.0f%8.0f%10.3E%7.3f%6.1f%7.0f%6.1f%4d%4d%4d%2d\n' \
                    %(cryr,crmnt,crdy,crhr,data[k,0],data[k,1],data[k,2],emis,data[k,4],data[k,5],data[k,6],data[k,7],data[k,8],data[k,9],data[k,10],data[k,11]))
        file.close()
        
    def run_STILT(self,j,n,k,spc,psdo):
        """This function reads in STILT footprints and returns hourly concentration data"""
    
        #read temporal profiles for the times within the footprint
        if psdo==0:
            dtkey=(self.timefinishkey-self.timestartkey).days*24+(self.timefinishkey-self.timestartkey).seconds/3600+1
        elif psdo==1:
            dtkey=int(self.dacycle['time.cycle'])*24
        idx=j+dtkey
        tempfile = os.path.join(self.model_settings['run.datadir'],'prior_temporal_%03d.nc'%n)
        tf = io.ct_read(tempfile, 'read')
        temp_profiles = []
        for s in range(self.nrcat):
            f = tf.get_variable(self.temporal_var[s])[idx-self.btime:idx]
            temp_profiles.append(f)
        tf.close()
        temp_profiles = np.array(temp_profiles)
               
        #read spatially distributed emissions calculated with the emission model
        emisfile = os.path.join(self.model_settings['run.datadir'],'prior_spatial_%03d.nc'%n) #format: species[SNAP,lon,lat]
        f = io.ct_read(emisfile,method='read')
        emis=f.get_variable(self.spname[spc])
        f.close()
        
        #find correct background concentration
        if self.bgswitch == 1:
            bgr = self.bg[spc,j]
        else:
            bgr=0.
                            
        #find correct footprint file
        tms=[]
        tmser=[]
        if self.lat[k]>0:
            lt='N'
        else:
            lt='S'
        if self.lon[k]>0:
            ln='E'
        else:
            ln='W'
        dum='%.2f'%abs(self.lat[k])
        dum2=dum.zfill(5)
        lt2=dum2+lt
        dum='%.2f'%abs(self.lon[k])
        dum2=dum.zfill(6)
        ln2=dum2+ln
        ht2='%05d'%self.hgt[k]
        fprint=os.path.join(self.model_settings['run.footdir'], 'foot%04dx%02dx%02dx%02dx%sx%sx%s.nc'%(self.datlist[j].year,self.datlist[j].month,self.datlist[j].day,self.datlist[j].hour,lt2,ln2,ht2))
        mf = io.ct_read(fprint,method='read')
        footprint=mf.get_variable('foot1')
        mf.close()
        
        # multiply footprint with emissions for each hour in the back trajectory
        for l in range(footprint.shape[0]):
            ft=footprint[l]
            dumvar=np.array(emis.shape[1]*[emis.shape[2]*[0]])
            for i in range(emis.shape[0]):
                emis_new=emis[i]*temp_profiles[i,l]
                dumvar=dumvar+emis_new
            dum=np.sum(ft*dumvar*self.cr[spc],axis=0)
            tms.append(dum)
        tms=np.array(tms)
        tmser=tms.sum()+bgr

        return tmser
            
    def validate_input(self):
        """ Make sure that data needed for the ObservationOperator (such as observation input lists, or parameter files)
            are present.
        """
        
        datadir = self.dacycle['dir.input']
        if not os.path.exists(datadir):
            msg = "The specified input directory for the OPS model to read from does not exist (%s), exiting..." % datadir 
            logging.error(msg)
            raise IOError, msg

        datafiles = os.listdir(datadir)

        obsfile = self.dacycle['ObsOperator.inputfile']

        if not os.path.exists(obsfile):
            msg = "The specified obs input file for the OPS model to read from does not exist (%s), exiting..." % obsfile  
            logging.error(msg)
            if not self.dacycle.has_key('forward.savestate.dir'): 
                raise IOError, msg

        for n in range(int(self.dacycle['da.optimizer.nmembers'])):
            paramfile = 'parameters.%03d.nc' % n
            if paramfile not in datafiles:
                msg = "The specified parameter input file for the OPS model to read from does not exist (%s), exiting..." % paramfile 
                logging.error(msg)
                raise IOError, msg
                
        self.ops_exec = self.model_settings['run.opsexec']
        if not os.path.exists(self.ops_exec):
            logging.error("Required OPS executable was not found %s" % self.ops_exec)
            logging.error("Please compile the model first")
            raise IOError


    def run_forecast_model(self,dacycle,psdo,adv):
        if psdo==0:
            self.prepare_run(psdo,adv)
            self.validate_input()
            self.run(dacycle,psdo)
            self.save_data()
        elif psdo==1:
            self.prepare_run(psdo,adv)
            self.run(dacycle,psdo)
            self.save_obs()

    def run(self,dacycle,psdo):
        """ This function calls the OPS and STILT functions for each locations, species, ensemble member and time step"""

        #set time control variables for this cycle
        if psdo==0:
            self.timestartkey = dtm.datetime(self.dacycle['time.sample.start'].year,self.dacycle['time.sample.start'].month,self.dacycle['time.sample.start'].day,self.dacycle['time.sample.start'].hour,0)
            self.timefinishkey = dtm.datetime(self.dacycle['time.sample.end'].year,self.dacycle['time.sample.end'].month,self.dacycle['time.sample.end'].day,self.dacycle['time.sample.end'].hour,0)
        elif psdo==1:
            self.timestartkey = dtm.datetime(self.dacycle['time.fxstart'].year,self.dacycle['time.fxstart'].month,self.dacycle['time.fxstart'].day,self.dacycle['time.fxstart'].hour,0)
            self.timefinishkey = dtm.datetime(self.dacycle['time.finish'].year,self.dacycle['time.finish'].month,self.dacycle['time.finish'].day,self.dacycle['time.finish'].hour,0)
        dt=dtm.timedelta(0,3600)
        dumdate=self.timestartkey
        self.datlist=[]
        while dumdate<(self.timefinishkey+dt):
            self.datlist.append(dumdate)
            dumdate=dumdate+dt
        
        #first remove old OPS output files and make sure output directories exist
        for s in range(self.nrspc):
            dirname = os.path.join(self.model_settings['run.opsoutdir'],'%03d'%(s+1))
            if os.path.isdir(dirname):
                filename = os.path.join(self.model_settings['run.opsoutdir'],'%03d/run%03d_999.plt'%(s+1,s+1))
                try:
                    os.remove(filename)
                except OSError:
                    pass
                for n in range(int(self.dacycle['da.optimizer.nmembers'])):
                    filename = os.path.join(self.model_settings['run.opsoutdir'],'%03d/run%03d_%03d.plt'%(s+1,s+1,n))
                    try:
                        os.remove(filename)
                    except OSError:
                        pass
            else:
                os.makedirs(dirname)
                
        #read in background concentrations
        if self.bgswitch == 1:
            bf = io.ct_read(self.bgfile, 'read')
            itimes = bf.get_variable('Times')
            bgtimes = array([dtm.datetime(int(''.join(d[:4])),int(''.join(d[5:7])),int(''.join(d[8:10])),int(''.join(d[11:13])),int(''.join(d[14:16]))) for d in itimes])
            idx1=np.where(bgtimes==self.timestartkey)[0]
            idx2=np.where(bgtimes==self.timefinishkey)[0]
            self.bg = []
            for s in range(self.nrspc):
                if psdo == 1:
                    varname = '%sbg_true'%(self.spname[s])
                elif psdo == 0:
                    #varname = '%sbg_true'%(self.spname[s])
                    varname = '%sbg_prior'%(self.spname[s])
                if varname in bf.variables.keys():
                    bgc = bf.get_variable(varname)[idx1[0]:idx2[0]+1]
                else:
                    bgc = np.zeros(idx2-idx1+1)
                self.bg.append(bgc)
            self.bg=np.array(self.bg)
            bf.close()

        if psdo==1: 
        #create pseudo-obs (full time series at once); for this we set n (ensemble member) to 999 as a flag
            self.conc_OPS=(self.nrloc*self.nrspc*len(self.datlist))*[0.]
            dums2_all=[]
            self.ids=[]
            self.sps=[]
            #OPS always has to start at 1:00 a.m.
            self.modelstart = dtm.datetime(self.timestartkey.year,self.timestartkey.month,self.timestartkey.day-1,1,0)
            self.timestart = self.modelstart
            self.modelend = self.timefinishkey
            for k in range(self.nrloc):
                for s in range(self.nrspc):
                    if k == 0 and len(self.ops_id)>0: #run OPS only for first location, as all locations are in the output (no need to run multiple times)
                        dum1=self.OPS_run(n=999,d=999,spc=s,psdo=psdo)
                    for d in range(len(self.datlist)):
                        ######do STILT area sources
                        dum2=self.run_STILT(j=d,n=999,k=k,spc=s,psdo=psdo)
                        dums2_all.append(dum2)
                        #other variables that need to be in observation files: species and index
                        self.ids.append(d+len(self.datlist)*s+len(self.datlist)*self.nrspc*k)
                        self.sps.append('%3s'%self.spname[s])
            self.mod_prior=np.array(self.conc_OPS)+np.array(dums2_all)
            logging.debug('Finished simulating pseudo time series')
        else: 
            ####create model simulations per time step using X hour footprints
            #### start with OPS
            totser=int(self.dacycle['da.optimizer.nmembers'])*[(self.nrloc*self.nrspc*len(self.datlist))*[0.]]
            self.conc_OPS=np.array(int(self.dacycle['da.optimizer.nmembers'])*[(self.nrloc*self.nrspc*len(self.datlist))*[0.]])
            if len(self.ops_id)>0:
                if self.dacycle['da.resources.parallel']=='parallel':
                    import multiprocessing
                    for s in range(self.nrspc):
                        for d in range(len(self.datlist)):
                            self.timestart = self.datlist[d] - dtm.timedelta(0,self.btime*3600)
                            if self.timestart.hour != 0:
                                self.modelstart = dtm.datetime(self.timestart.year,self.timestart.month,self.timestart.day,1,0)
                            else:
                                self.modelstart = dtm.datetime(self.timestart.year,self.timestart.month,self.timestart.day-1,1,0)
                            self.modelend = self.datlist[d]
                            p = multiprocessing.Pool(int(self.dacycle['da.resources.ntasks']))
                            args=[]
                            ##parallelize OPS runs for ensemble members
                            for n in range(int(self.dacycle['da.optimizer.nmembers'])):
                                dum=[self,n,d,s,psdo]
                                args.append(dum)
                            data = p.map(dumfunc,args)
                            for m in range(len(data)):
                                for t in range(self.nrloc):
                                    self.conc_OPS[m,d+len(self.datlist)*s+len(self.datlist)*self.nrspc*t]=data[m][t]
                            p.close()
                            p.join()
                else:
                    for n in range(int(self.dacycle['da.optimizer.nmembers'])):
                        for s in range(self.nrspc):
                            for d in range(len(self.datlist)):
                                self.timestart = self.datlist[d] - dtm.timedelta(0,self.btime*3600)
                                if self.timestart.hour != 0:
                                    self.modelstart = dtm.datetime(self.timestart.year,self.timestart.month,self.timestart.day,1,0)
                                else:
                                    self.modelstart = dtm.datetime(self.timestart.year,self.timestart.month,self.timestart.day-1,1,0)
                                self.modelend = self.datlist[d]
                                dum1=self.OPS_run(n=n,d=d,spc=s,psdo=psdo)
                                for t in range(self.nrloc):
                                    self.conc_OPS[n,d+len(self.datlist)*s+len(self.datlist)*self.nrspc*t]=dum1[t]
            #### then do STILT
            for n in range(int(self.dacycle['da.optimizer.nmembers'])):
                conc_STILT=[]
                for k in range(self.nrloc):
                    for s in range(self.nrspc):
                        for d in range(len(self.datlist)):
                            dum2=self.run_STILT(j=d,n=n,k=k,spc=s,psdo=psdo)
                            conc_STILT.append(dum2)
                totser[n]=np.array(self.conc_OPS[n])+np.array(conc_STILT)
                # totser[n]=np.array(conc_STILT)
            self.mod=np.array(totser).T
            logging.debug('Finished doing simulations per back trajectory time')

    def save_data(self):
        """ Write the data that is needed for a restart or recovery of the Observation Operator to the save directory """

        import da.tools.io4 as io
        
        f = io.CT_CDF(self.simulated_file, method='create')
    	logging.debug('Creating new simulated observation file in ObservationOperator (%s)' % self.simulated_file)
	
        dimid = f.createDimension('obs_num', size=None)
        dimid = ('obs_num',)
        savedict = io.std_savedict.copy() 
        savedict['name'] = "obs_num"
        savedict['dtype'] = "int"
        savedict['long_name'] = "Unique_Dataset_observation_index_number"
        savedict['units'] = ""
        savedict['dims'] = dimid
        savedict['comment'] = "Unique index number within this dataset ranging from 0 to UNLIMITED."
        f.add_data(savedict,nsets=0)

        dimmember = f.createDimension('nmembers', size=self.forecast_nmembers)
        dimmember = ('nmembers',)
        savedict = io.std_savedict.copy() 
        savedict['name'] = "model"
        savedict['dtype'] = "float"
        savedict['long_name'] = "mole_fraction_of_trace_gas_in_air"
        savedict['units'] = "mol tracer (mol air)^-1"
        savedict['dims'] = dimid + dimmember
        savedict['comment'] = "Simulated model value"
        f.add_data(savedict,nsets=0)
        
        f_in = io.ct_read(self.dacycle['ObsOperator.inputfile'],method='read') 
        ids = f_in.get_variable('obs_num')
        
        for i,data in enumerate(zip(ids,self.mod)):
            f.variables['obs_num'][i] = data[0]		
            f.variables['model'][i,:] = data[1]
            dum=f.variables['model'][:]
            
        f.close()
        f_in.close()
        
    def save_obs(self):
        """Write pseudo-observations to file"""

        ct1=0
        for k in range(self.nrloc*self.nrspc):
            newfile = os.path.join(self.obsdir,self.obsnc[k]+'.nc')
            f = io.CT_CDF(newfile, method='create')
            logging.debug('Creating new pseudo observation file in ObservationOperator (%s)' % newfile)
            
            dimid = f.add_dim('Time',len(self.datlist))
            ln = len(self.datlist)
            str19 = f.add_dim('strlen',19)
            str3 = f.add_dim('strlen2',3)

            data=[]
            for it,t in enumerate(self.datlist):
                data.append(list(t.isoformat().replace('T','_')))
            
            savedict = io.std_savedict.copy() 
            savedict['dtype'] = "char"
            savedict['name'] = "Times"
            savedict['units'] = ""
            savedict['dims'] = dimid + str19
            savedict['values'] = data
            f.add_data(savedict)
            
            data = ln*[self.lat[int(k/self.nrspc)]]
            savedict = io.std_savedict.copy() 
            savedict['name'] = "lat"
            savedict['units'] = "degrees_north"
            savedict['dims'] = dimid
            savedict['values'] = data
            f.add_data(savedict)
            
            data = ln*[self.lon[int(k/self.nrspc)]]
            savedict = io.std_savedict.copy() 
            savedict['name'] = "lon"
            savedict['units'] = "degrees_east"
            savedict['dims'] = dimid
            savedict['values'] = data
            f.add_data(savedict)
            
            data = ln*[self.hgt[int(k/self.nrspc)]]
            savedict = io.std_savedict.copy() 
            savedict['name'] = "alt"
            savedict['units'] = "m above ground"
            savedict['dims'] = dimid
            savedict['values'] = data
            f.add_data(savedict)
            
            savedict = io.std_savedict.copy() 
            savedict['name'] = "obs"
            savedict['units'] = "ppm or ppb"
            savedict['dims'] = dimid
            savedict['values'] = self.mod_prior[ct1:ct1+ln]
            f.add_data(savedict)
            
            savedict = io.std_savedict.copy() 
            savedict['dtype'] = "char"
            savedict['name'] = "species"
            savedict['units'] = "observed species"
            savedict['dims'] = dimid + str3
            savedict['values'] = self.sps[ct1:ct1+ln]
            f.add_data(savedict)
                
                
            savedict = io.std_savedict.copy() 
            savedict['dtype'] = "int"
            savedict['name'] = "ids"
            savedict['units'] = "unique observation identification number"
            savedict['dims'] = dimid
            savedict['values'] = self.ids[ct1:ct1+ln]
            f.add_data(savedict)
            

            f.close()
            
            ct1=ct1+ln
            
            logging.debug("Successfully wrote data to obs file %s"%newfile)
            

################### End Class STILT ###################


if __name__ == "__main__":
    pass


